-- Mutt - Macro Update Target Tool
-- Tuill of Greymane
-- Revisions:
-- 0.1 - Initial copy of Telegraph (by Tuill) source
-- 0.2 - First running test
-- 0.3 - Shared as private beta
-- 0.32 - Undid move of combat lockdown test
-- 0.40 - Upload as alpha to WowAce

-- Many sources of inspiration (== blatant copy/pastes)
-- All comments by Tuill
-- I recommend a Lua-aware editor like SciTE that provides syntactic highlighting.

-- No global for now
-- Mutt = LibStub("AceAddon-3.0"):NewAddon("Mutt", "AceConsole-3.0")

-- local scope identifier for performance and template-ability
local ourAddon = LibStub("AceAddon-3.0"):NewAddon("Mutt", "AceConsole-3.0")

-- local scope identifiers for util functions
local strlower = strlower
local tonumber = tonumber
local string_gsub = string.gsub
local table_insert = table.insert

-- Fetch version & notes from TOC file
local ourVersion = GetAddOnMetadata("Mutt", "Version")
local ourNotes = GetAddOnMetadata("Mutt", "Notes")

-- Multi-line string for the help text, color codes inline
local helpUsing = [[Examples:

|cff999999# Assuming you have a macro named "hlight" that looks like:|r
/cast [target=raid4] Holy Light

|cff999999# ...and that you're targeting the player in position raid23, then|r
/mutt hlight

|cff999999# ...would patch the hlight macro to look like:|r
/cast [target=raid23] Holy Light

|cff999999# Mutt observes macro options, so you can even do:|r
/mutt [button:3] hlight
/cast [nobutton:3,target=raid2] Holy Light

|cff999999# ...to have a macro patch itself(!)|r

|cffCC8888# WARNING!|r
|cff999999
# Be advised! If you try this macro-patching-itself trick,
# make absolutely certain that the Mutt slash-command is the
# first line of the macro, and that you use macro options to
# make it mutually-exclusive from the rest of your macro
# or else your results will be unpredictable.|r

modifiers:
|cff999999
# Mutt modifiers can be added after the name of the macro. If
# no modifiers are provided, Mutt will patch the first target=
# that it finds (unless it's part of a /mutt command, see
# below) and set it to the raid position of the currently
# targeted raid member.

# Adding numbers after the name of the macro tells Mutt to
# change those instances of target= in the macro, so:|r
/mutt hlight 2
|cff999999
# ...would change the second target= in your macro. In case
# you want to change the last target= and don't want to count
# how many come before it, Mutt understands negative numbers
# to mean that you want to count backward from the end, so:|r
/mutt hlight -2
|cff999999
# ...would be the next-to-last target= in your macro.

# Placing|r all |cff999999as an modifier in your Mutt command tells Mutt to
# change all instances of target= in your macro to what Mutt
# received as the current target (more on targeting below).

# By default Mutt won't count or change any target= that are
# part of a Mutt command in a macro. You can change this
# behavior if you wish by including the|r mutt |cff999999modifier, and the
# target= in your Mutt commands will be treated like those in
# the rest of your macro. Please note the |cffCC8888warning |cff999999above that
# unpredictable/undesired behavior may result from this.

# The raid position that Mutt writes in your updated macro
# is determined by your current target or by macro options
# in your Mutt command if you provide them, so if you used:|r
/mutt [target=focus] hlight
|cff999999
# ...and your focus target was the player in raid position 14,
then raid14 would be the new target written in your macro.
|r
]]

-- Ace3 options table
-- RegisterOptionsTable can take a function ref for the options arg
-- Taking advantage of this in case we decide to dynamically adjust
-- at use-time
local function ourOptions()
  local options = { 
   type = 'group',
   args = {
	general = {
	  type = 'group',
	  name = "Settings",
	  args = {
			header1 =
			{
				order = 1,
				type = "header",
				name = "",
			},
			version =
			{
				order = 2,
				type = "description",
				name = "Version " .. ourVersion .. "\n",
			},
			about =
			{
				order = 3,
				type = "description",
				name = ourNotes.."\n",
			},
			about2 =
			{
				order = 5,
				type = "description",
				name = helpUsing
			},
		}
	} -- end using
   } -- top args
  } -- end table
 return options
end

function ourAddon:OnInitialize()

	local ourConfig = LibStub("AceConfig-3.0")
	local ourConfigDialog = LibStub("AceConfigDialog-3.0") -- For snapping into the ESC menu addons list

	ourConfig:RegisterOptionsTable("Mutt", ourOptions)

	self.optionsFrames = {}
	self.optionsFrames.general = ourConfigDialog:AddToBlizOptions("Mutt", "Mutt", nil, "general")

	-- Create slash commands
	self:RegisterChatCommand("mutt", "SlashHandler")
--  self:RegisterChatCommand("muttdebug", "MDebugHandler")
end


function ourAddon:SlashHandler(input)
  -- Show addon config dialog as help if no args, we verify we're not in combat
  -- so it should be too annoying (better than just a blurt in the chat pane...)
  if input == "" then
    if InCombatLockdown() then
	  self:Print("In combat, declining to show Mutt help dialog.")
	else
	  InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.general)
	end
  else
    self:TargetUpdate(0, input)
  end
end

-- Functions defined in a do block so we could have pseudo-static variables
do
   
  function ourAddon:TargetUpdate(flag, args)
  -- Declare variables
    local ourArgs, alterMutt, targetCount, ourTarget, macroIndex
	local inaParty, inaRaid, groupMember, totalMembers
	local groupName, targetingSelf, targetingSelfPet, petTag
	
	-- Initialize and object we'll use to represent this /mutt command
	local muttCall = { macroMutts = 0, groupNumber = nil, countTargets = 0, alterAll = false, ourMacro = nil, ourMuttlines = {}, alterTargets = {}, mt = {} }
	setmetatable(muttCall, muttCall.mt)
	
	-- Extract macro to patch and target
    ourArgs, ourTarget = SecureCmdOptionParse(args)
	if not ourArgs then
	  -- No macro, so stop here
	  -- If we're calling /mutt with macro options, ourArgs will
	  -- be nil in cases where there's no "true" option, so correct 
	  -- behavior here is to silently return (do nothing).
	  return
	end

    if InCombatLockdown() then
	  self:Print("Can't adjust macros in combat.")
 	  return
    end
	
	-- Step over args provided, pick out the macro name first
	for step in ourArgs:gmatch("%S+") do
	  if not muttCall.ourMacro then
	    muttCall.ourMacro = step
	  else
	    step = strlower(step)
	  end
	  if tonumber(step) then
	    table_insert(muttCall.alterTargets, tonumber(step))
	  elseif step == "all" then
	    muttCall.alterAll = true
	  elseif step == "mutt" then
	    alterMutt = true
	  end
	end

    -- Target not provided or not understood, default to player's target
    if not (ourTarget) then
      ourTarget = "target"
    end
    if not (UnitExists(ourTarget)) then
	  self:Print("Nothing targeted or can't determine target, no update to macro "..muttCall.ourMacro)
      return
    end
	macroIndex = GetMacroIndexByName(muttCall.ourMacro)
	if macroIndex == 0 then
	  -- Not a WoW macro, so we can't deal with it...
	  self:Print("Can't find macro "..muttCall.ourMacro.." (is it an add-on macro not a WoW macro..?)")
	  return
	end
	
	-- Addon Lesson #102:
    --	A call to UnitInParty("player") always returns true, 
	--  so we test for party1 to see if we're in a party
	inaParty = UnitExists("party1")
	inaRaid = UnitInRaid("player")
	targetingSelf = UnitIsUnit(ourTarget, "player")
	targetingSelfPet = UnitIsUnit(ourTarget, "playerpet")
	
	  -- Choose biggest grouping if none specified
    if inaRaid then
	    groupMember = UnitPlayerOrPetInRaid(ourTarget)
		totalMembers = GetNumRaidMembers()
		groupName = "raid"
	elseif inaParty then
	    groupMember = UnitPlayerOrPetInParty(ourTarget)
		totalMembers = GetNumPartyMembers()
		groupName = "party"
	else
	    self:Print("Not in a group, can only reassign target to group positions.")
	    return
	end
	
	if (groupMember or targetingSelfPet) and (not UnitIsPlayer(ourTarget)) then
	  -- targeting a pet
	  petTag = "pet"
	else
	  petTag = ""
	end
	
	-- If we're here we have an existing target
	if groupMember or targetingSelf or targetingSelfPet then
	  local targetGUID = UnitGUID(ourTarget)
	  for step = 1, totalMembers do
	    -- This builds a name like 'party2' or 'raid4pet'
	    muttCall.groupNumber = groupName..step..petTag
	    if UnitGUID(muttCall.groupNumber) == targetGUID then
		  -- Found the target's group name, no need to search more
		  break
		else
		  muttCall.groupNumber = nil
		end
	  end
	  -- Special case in a party group for player and the player's pet
      if (not muttCall.groupNumber) and (targetingSelf or targetingSelfPet) then
		  self:Print("You and your pet have no party# ID, defaulting to 'player"..petTag.."'")
		  muttCall.groupNumber = "player"..petTag
	  end

	  if not muttCall.groupNumber then
	    -- Bad!!
	    self:Print("Didn't get a valid group position for unit.")
	  else
	    local initial, targetEquals, remainder
		
		-- Get current info from provided macro
		local ourName, ourTexture, ourMacroBody, isLocal  = GetMacroInfo(macroIndex)

		-- Munge /mutt lines unless option is set to include them in count
		if not alterMutt then
		  -- We get string.gsub for free, so make the most of it
		  -- Read the Lua docs for string.gsub and the index event if you don't understand these lines
		  muttCall.mt.__index = ourAddon.muttMunge
		  ourMacroBody = string_gsub(ourMacroBody, "/mutt%s+([^\n]+)", muttCall)
		end
		
		-- Calling string.gsub for the side-effect of getting a match count
		_, targetCount = string_gsub(ourMacroBody, "(target=)(%w+)", {})
		
		if targetCount then
		  if (#muttCall.alterTargets == 0) and (not muttCall.alterAll) then
		    -- No explicit target= specified, default to 1st
			table_insert(muttCall.alterTargets, 1)
		  end

		  -- Step over targets to change, reconciling any negative numbers
		  local tempTable = {}
		  for ourKey, ourValue in pairs(muttCall.alterTargets) do
		    if ourValue < 0 then
			  ourValue = targetCount + 1 + ourValue
			end
			-- Eliminate out-of-bounds target instances,
			-- Build a working table with keys for valid target instances
			if (ourValue < 1) or (ourValue > targetCount) then
			  self:Print("Invalid target instance of "..ourValue.." provided, discarding...")
			else
			  tempTable[ourValue] = true
			end
		  end
		  muttCall.alterTargets = tempTable

		  -- Do the substitution
		  muttCall.mt.__index = ourAddon.muttTargetSub
		  ourMacroBody = string_gsub(ourMacroBody, "target=(%w+)", muttCall)
		  
		  -- De-munge /mutt lines if we did so
		  if not alterMutt then
		    ourMacroBody = string_gsub(ourMacroBody, "/mutt%s+([^\n]+)", muttCall.ourMuttlines)
		  end

		  -- Save our work
		  EditMacro(macroIndex, ourName, ourTexture, ourMacroBody, isLocal)
		else
		  self:Print("Couldn't find a predefined target in macro "..ourName)
		end
	  end
	else
	  -- We're targeting something outside our group
	  self:Print("We're not in a group with "..ourTarget)
	  return
	end
  end

  function ourAddon.muttMunge(slashCall, muttTarget)
	slashCall.macroMutts = slashCall.macroMutts + 1
	muttKey = "###"..slashCall.macroMutts
	slashCall.ourMuttlines[muttKey] = "/mutt "..muttTarget
	return "/mutt "..muttKey
  end
  
  function ourAddon.muttTargetSub(slashCall, prevTarget)
    slashCall.countTargets = slashCall.countTargets + 1
	local foo = slashCall.countTargets
	if (slashCall.alterTargets[foo]) or (slashCall.alterAll) then
	  ourAddon:Print("|cffccccccPatching target instance |r"..slashCall.countTargets.."|cffcccccc in macro |r"..slashCall.ourMacro.."|cffcccccc from |rtarget="..prevTarget.."|cffcccccc to |rtarget="..slashCall.groupNumber)
	  return "target="..slashCall.groupNumber
	else
	  return nil
	end
  end
  
end
